@@ -27,12 +27,18 @@ def setup(app: Sphinx) -> Dict[str, Any]:
     * connect events to functions
     """
     app.add_config_value("bibtex_default_style", "alpha", "html")
+    app.add_config_value("bibtex_tooltips", True, "html")
+    app.add_config_value("bibtex_tooltips_style", "", "html")
     app.add_config_value("bibtex_bibfiles", None, "html")
     app.add_config_value("bibtex_encoding", "utf-8-sig", "html")
     app.add_config_value("bibtex_bibliography_header", "", "html")
     app.add_config_value("bibtex_footbibliography_header", "", "html")
     app.add_config_value("bibtex_reference_style", "label", "env")
     app.add_config_value("bibtex_foot_reference_style", "foot", "env")
+    app.add_config_value("bibtex_cite_id", "", "html")
+    app.add_config_value("bibtex_footcite_id", "", "html")
+    app.add_config_value("bibtex_bibliography_id", "", "html")
+    app.add_config_value("bibtex_footbibliography_id", "", "html")
     app.add_domain(BibtexDomain)
     app.add_directive("bibliography", BibliographyDirective)
     app.add_role("cite", CiteRole())
@@ -45,8 +51,8 @@ def setup(app: Sphinx) -> Dict[str, Any]:
     app.add_role("footcite", FootCiteRole())
 
     return {
-        'version': '2.3.0',
-        'env_version': 7,
+        'version': '2.4.2',
+        'env_version': 9,
         'parallel_read_safe': True,
         'parallel_write_safe': True,
         }
