@@ -5,19 +5,13 @@
     .. autoclass:: Citation
         :members:
 
-    .. autoclass:: SphinxReferenceInfo
-        :members:
-
-    .. autoclass:: SphinxReferenceText
-        :members:
-
     .. autoclass:: BibtexDomain
         :members:
 """
 
 import ast
 from typing import TYPE_CHECKING
-from typing import List, Dict, NamedTuple, cast, Iterable, Tuple, Set
+from typing import List, Dict, NamedTuple, cast, Iterable, Tuple, Set, Optional
 
 import docutils.frontend
 import docutils.nodes
@@ -36,17 +30,13 @@ import re
 from sphinx.domains import Domain, ObjType
 from sphinx.errors import ExtensionError
 from sphinx.locale import _
-from sphinx.util.nodes import make_refnode
 
 from .roles import CiteRole
-from .bibfile import BibFile, normpath_filename, process_bibfile
-from .nodes import raw_latex
-from .style.referencing import (
-    BaseReferenceText, BaseReferenceStyle, format_references
-)
+from .bibfile import normpath_filename, process_bibdata, BibData
+from .style.referencing import BaseReferenceStyle, format_references
+from .style.template import SphinxReferenceInfo
 
 if TYPE_CHECKING:
-    from pybtex.backends import BaseBackend
     from pybtex.database import Entry
     from pybtex.style.formatting import BaseStyle
     from sphinx.addnodes import pending_xref
@@ -200,7 +190,9 @@ class _FilterVisitor(ast.NodeVisitor):
 def get_docnames(env):
     """Get document names in order."""
     rel = env.collect_relations()
-    docname = env.config.master_doc
+    docname = (
+        env.config.master_doc if sphinx.version_info < (4, 0)
+        else env.config.root_doc)
     docnames = set()
     while docname is not None:
         docnames.add(docname)
@@ -217,49 +209,8 @@ class Citation(NamedTuple):
     bibliography_key: "BibliographyKey"  #: Key of its bibliography directive.
     key: str                             #: Key (with prefix).
     entry: "Entry"                       #: Entry from pybtex.
-    formatted_entry: "FormattedEntry"    #: Entry as formatted by pybtex.
-
-
-class SphinxReferenceInfo(NamedTuple):
-    """Tuple containing reference info to enable sphinx to resolve a reference
-    to a citation.
-    """
-    builder: "Builder"  #: The Sphinx builder.
-    fromdocname: str    #: Document name of the citation reference.
-    todocname: str      #: Document name of the bibliography.
-    citation_id: str    #: Unique id of the citation within the bibliography.
-    title: str          #: Title attribute for reference node.
-
-
-class SphinxReferenceText(BaseReferenceText[SphinxReferenceInfo]):
-    """Pybtex rich text class for citation references with the docutils
-    backend, for use with :class:`SphinxReferenceInfo`.
-    """
-
-    def render(self, backend: "BaseBackend"):
-        assert isinstance(backend, pybtex_docutils.Backend), \
-               "SphinxReferenceText only supports the docutils backend"
-        info = self.info[0]
-        if info.builder.name == 'latex':
-            key = f'cite.{info.todocname}:{info.citation_id}'
-            return (
-                [raw_latex(f'\\hyperlink{{{key}}}{{')]
-                + super().render(backend)
-                + [raw_latex('}')]
-            )
-        else:
-            children = super().render(backend)
-            # make_refnode only takes a single child
-            refnode = make_refnode(
-                builder=info.builder,
-                fromdocname=info.fromdocname,
-                todocname=info.todocname,
-                targetid=info.citation_id,
-                child=children[0],
-                title=info.title,
-            )
-            refnode.extend(children[1:])  # type: ignore
-            return [refnode]
+    formatted_entry: "FormattedEntry"    #: Formatted entry for bibliography.
+    tooltip_entry: Optional["FormattedEntry"]  #: Formatted entry for tooltip.
 
 
 def env_updated(app: "Sphinx", env: "BuildEnvironment") -> Iterable[str]:
@@ -269,8 +220,9 @@ def env_updated(app: "Sphinx", env: "BuildEnvironment") -> Iterable[str]:
 
 def parse_header(header: str, source_path: str):
     parser = docutils.parsers.rst.Parser()
+    # note: types stub for docutils doesn't know about components argument
     settings = docutils.frontend.OptionParser(
-        components=(docutils.parsers.rst.Parser,)
+        components=(docutils.parsers.rst.Parser,)  # type: ignore
     ).get_default_values()
     document = docutils.utils.new_document(source_path, settings)
     parser.parse(header, document)
@@ -282,11 +234,13 @@ class BibtexDomain(Domain):
 
     name = 'cite'
     label = 'BibTeX Citations'
-    data_version = 3
+    data_version = 4
     initial_data = dict(
-        bibfiles={},
-        bibliography_header=docutils.nodes.paragraph(),
-        footbibliography_header=docutils.nodes.paragraph(),
+        bibdata=BibData(
+            encoding='',
+            bibfiles={},
+            data=pybtex.database.BibliographyData()),
+        bibliography_header=docutils.nodes.container(),
         bibliographies={},
         citations=[],
         citation_refs=[],
@@ -295,11 +249,9 @@ class BibtexDomain(Domain):
     reference_style: BaseReferenceStyle
 
     @property
-    def bibfiles(self) -> Dict[str, BibFile]:
-        """Map each bib filename to some information about the file (including
-        the parsed data).
-        """
-        return self.data['bibfiles']
+    def bibdata(self) -> BibData:
+        """Information about the bibliography files."""
+        return self.data['bibdata']
 
     @property
     def bibliography_header(self) -> docutils.nodes.Element:
@@ -341,14 +293,15 @@ class BibtexDomain(Domain):
             raise ExtensionError(
                 "You must configure the bibtex_bibfiles setting")
         # update bib file information in the cache
-        for bibfile in env.app.config.bibtex_bibfiles:
-            process_bibfile(
-                self.bibfiles, normpath_filename(env, "/" + bibfile),
-                env.app.config.bibtex_encoding)
+        bibfiles = [
+            normpath_filename(env, "/" + bibfile)
+            for bibfile in env.app.config.bibtex_bibfiles]
+        self.data['bibdata'] = process_bibdata(
+            self.bibdata, bibfiles, env.app.config.bibtex_encoding)
         # parse bibliography header
         header = getattr(env.app.config, "bibtex_bibliography_header")
         if header:
-            self.data["bibliography_header"] = \
+            self.data["bibliography_header"] += \
                 parse_header(header, "bibliography_header")
 
     def clear_doc(self, docname: str) -> None:
@@ -383,19 +336,24 @@ class BibtexDomain(Domain):
         used_keys: Set[str] = set()
         used_labels: Dict[str, str] = {}
         for bibliography_key, bibliography in self.bibliographies.items():
-            for entry, formatted_entry in self.get_formatted_entries(
-                    bibliography_key, docnames):
+            for entry, formatted_entry, tooltip_entry in \
+                    self.get_formatted_entries(
+                        bibliography_key, docnames,
+                        self.env.app.config.bibtex_tooltips,
+                        self.env.app.config.bibtex_tooltips_style):
                 key = bibliography.keyprefix + formatted_entry.key
                 if bibliography.list_ == 'citation' and key in used_keys:
                     logger.warning(
                         'duplicate citation for key "%s"' % key,
-                        location=(bibliography_key.docname, bibliography.line))
+                        location=(bibliography_key.docname, bibliography.line),
+                        type="bibtex", subtype="duplicate_citation")
                 self.citations.append(Citation(
                     citation_id=bibliography.citation_nodes[key]['ids'][0],
                     bibliography_key=bibliography_key,
                     key=key,
                     entry=entry,
                     formatted_entry=formatted_entry,
+                    tooltip_entry=tooltip_entry,
                 ))
                 if bibliography.list_ == 'citation':
                     used_keys.add(key)
@@ -409,7 +367,8 @@ class BibtexDomain(Domain):
                                 formatted_entry.label,
                                 used_labels[formatted_entry.label], key),
                             location=(bibliography_key.docname,
-                                      bibliography.line))
+                                      bibliography.line),
+                            type="bibtex", subtype="duplicate_label")
         return []  # expects list of updated docnames
 
     def resolve_xref(self, env: "BuildEnvironment", fromdocname: str,
@@ -425,7 +384,8 @@ class BibtexDomain(Domain):
         for key in keys:
             if key not in citations:
                 logger.warning('could not find bibtex key "%s"' % key,
-                               location=node)
+                               location=node, type="bibtex",
+                               subtype="key_not_found")
         plaintext = pybtex.plugin.find_plugin('pybtex.backends', 'plaintext')()
         references = [
             (citation.entry, citation.formatted_entry, SphinxReferenceInfo(
@@ -433,11 +393,14 @@ class BibtexDomain(Domain):
                 fromdocname=fromdocname,
                 todocname=citation.bibliography_key.docname,
                 citation_id=citation.citation_id,
-                title=citation.formatted_entry.text.render(plaintext)))
+                title=(
+                    citation.tooltip_entry.text.render(plaintext)
+                    if citation.tooltip_entry else None
+                )
+            ))
             for citation in citations.values()]
-        formatted_references = \
-            format_references(
-                self.reference_style, SphinxReferenceText, typ, references)
+        formatted_references = format_references(
+            self.reference_style, typ, references)
         result_node = docutils.nodes.inline(rawsource=target)
         result_node += formatted_references.render(self.backend)
         return result_node
@@ -476,8 +439,8 @@ class BibtexDomain(Domain):
         in order of appearance in the bib files.
         """
         for bibfile in bibfiles:
-            for entry in self.bibfiles[bibfile].data.entries.values():
-                yield entry
+            for key in self.bibdata.bibfiles[bibfile].keys:
+                yield self.bibdata.data.entries[key]
 
     def get_filtered_entries(
             self, bibliography_key: "BibliographyKey"
@@ -502,7 +465,8 @@ class BibtexDomain(Domain):
             except ValueError as err:
                 logger.warning(
                     "syntax error in :filter: expression; %s" % err,
-                    location=(bibliography_key.docname, bibliography.line))
+                    location=(bibliography_key.docname, bibliography.line),
+                    type="bibtex", subtype="filter_syntax_error")
                 # recover by falling back to the default
                 success = bool(cited_docnames)
             if success or entry.key in bibliography.keys:
@@ -526,17 +490,23 @@ class BibtexDomain(Domain):
             yield key, entry
 
     def get_formatted_entries(
-            self, bibliography_key: "BibliographyKey", docnames: List[str]
-            ) -> Iterable[Tuple["Entry", "FormattedEntry"]]:
+            self, bibliography_key: "BibliographyKey", docnames: List[str],
+            tooltips: bool, tooltips_style: str
+            ) -> Iterable[Tuple["Entry", "FormattedEntry",
+                                Optional["FormattedEntry"]]]:
         """Get sorted bibliography entries along with their pybtex labels,
         with additional sorting and formatting applied from the pybtex style.
         """
         bibliography = self.bibliographies[bibliography_key]
         entries = dict(
             self.get_sorted_entries(bibliography_key, docnames))
-        style = cast("BaseStyle", pybtex.plugin.find_plugin(
+        style: BaseStyle = cast("BaseStyle", pybtex.plugin.find_plugin(
             'pybtex.style.formatting', bibliography.style)())
-        sorted_entries = style.sort(entries.values())
+        style2: Optional[BaseStyle] = (
+            cast("BaseStyle", pybtex.plugin.find_plugin(
+                'pybtex.style.formatting', tooltips_style)())
+            if tooltips_style else style) if tooltips else None
+        sorted_entries: Iterable[Entry] = style.sort(entries.values())
         labels = style.format_labels(sorted_entries)
         for label, entry in zip(labels, sorted_entries):
             try:
@@ -544,13 +514,16 @@ class BibtexDomain(Domain):
                     entry,
                     style.format_entry(
                         bibliography.labelprefix + label, entry),
+                    style2.format_entry(
+                        bibliography.labelprefix + label, entry)
+                    if style2 else None,
                 )
             except FieldIsMissing as exc:
                 logger.warning(
                     str(exc),
-                    location=(bibliography_key.docname, bibliography.line))
-                yield(
-                    entry,
-                    FormattedEntry(entry.key, Tag('b', str(exc)),
-                                   bibliography.labelprefix + label)
-                )
+                    location=(bibliography_key.docname, bibliography.line),
+                    type="bibtex", subtype="missing_field")
+                formatted_error_entry = FormattedEntry(
+                    entry.key, Tag('b', str(exc)),
+                    bibliography.labelprefix + label)
+                yield entry, formatted_error_entry, None
